@@ -17,11 +17,14 @@ BEGIN
 
     sub HANDLE    () { 16 | 32 }
     sub BOOLEAN   () { 1 | 256 }
+
+    my $val = $ENV{PERL_NO_VALIDATION} ? 1 : 0;
+    eval "sub NO_VALIDATE () { $val }";
 }
 
 require Exporter;
 
-use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS %OPTIONS $called $options);
+use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS %OPTIONS $options);
 @ISA = qw(Exporter);
 
 my %tags =
@@ -39,41 +42,75 @@ my %tags =
 @EXPORT_OK = ( @{ $EXPORT_TAGS{all} }, 'set_options' );
 @EXPORT = qw( validate validate_pos );
 
-$VERSION = '0.18';
+$VERSION = '0.22';
+
+=pod
+
+=begin internals
+
+Various internals notes (for me and any future readers of this
+monstrosity):
+
+- A lot of the weirdness is _intentional_, because it optimizes for
+  the _success_ case.  It does not really matter how slow the code is
+  after it enters a path that leads to reporting failure.  But the
+  "success" path should be as fast as possible.
+
+-- We only calculate $called as needed for this reason, even though it
+   means copying code all over.
+
+- All the validation routines need to be careful never to alter the
+  references that are passed.
+
+-- The code assumes that _most_ callers will not be using the
+   skip_leading or ignore_case features.  In order to not alter the
+   references passed in, we copy them wholesale when normalize them to
+   make these features work.  This is slower but lets us be faster
+   when not using them.
+
+=end
+
+=cut
+
 
 # Matt Sergeant came up with this prototype, which slickly takes the
 # first array (which should be the caller's @_), and makes it a
 # reference.  Everything after is the parameters for validation.
 sub validate_pos (\@@)
 {
+    return if NO_VALIDATE && ! defined wantarray;
+
     my $p = shift;
+
     my @specs = @_;
 
     my @p = @$p;
-    if ( $ENV{PERL_NO_VALIDATION} )
+    if ( NO_VALIDATE )
     {
-	foreach my $x (0..$#specs)
+        # if the spec is bigger that's where we can start adding
+        # defaults
+        for ( my $x = $#p + 1; $x <= $#specs; $x++ )
 	{
-	    if ( $x > $#p )
-	    {
-		$p[$x] =
-                    $specs[$x]->{default}
-                        if ref $specs[$x] && exists $specs[$x]->{default};
-	    }
+            $p[$x] =
+                $specs[$x]->{default}
+                    if ref $specs[$x] && exists $specs[$x]->{default};
 	}
 
-	return @p;
+	return wantarray ? @p : \@p;
     }
 
     # I'm too lazy to pass these around all over the place.
-    local $options ||= _get_options( (caller(0))[0] ) unless defined $options;
-    local $called  = (caller( $options->{stack_skip} ))[3] unless defined $called;
+    local $options ||= _get_options( (caller(0))[0] )
+        unless defined $options;
 
     my $min = 0;
 
     while (1)
     {
-	last if _is_optional( $specs[$min] );
+        last unless ( ref $specs[$min] ?
+                      ! ( exists $specs[$min]->{default} || $specs[$min]->{optional} ) :
+                      $specs[$min] );
+
 	$min++;
     }
 
@@ -90,6 +127,8 @@ sub validate_pos (\@@)
 	my $val = $options->{allow_extra} ? $min : $max;
 	$minmax .= $val != 1 ? ' were' : ' was';
 
+        my $called = _get_called();
+
 	$options->{on_fail}->
             ( "$actual parameter" .
               ($actual != 1 ? 's' : '') .
@@ -102,71 +141,107 @@ sub validate_pos (\@@)
     foreach ( 0..$bigger )
     {
 	my $spec = $specs[$_];
-	if ( $_ <= $#p )
-	{
-	    _validate_one_param( $p[$_], $spec, "Parameter #" . ($_ + 1) )
-		if ref $spec;
-	}
 
 	next unless ref $spec;
 
+	if ( $_ <= $#p )
+	{
+	    _validate_one_param( $p[$_], $spec, "Parameter #" . ($_ + 1) );
+	}
+
 	$p[$_] = $spec->{default} if $_ > $#p && exists $spec->{default};
     }
 
-    return @p if wantarray;
+    return wantarray ? @p : \@p;
 }
 
 sub validate (\@$)
 {
-    my $p = shift;
-    my %specs = %{ shift() };
+    return if NO_VALIDATE && ! defined wantarray;
+
+    my $p = $_[0];
+
+    my $specs = $_[1];
 
     local $options = _get_options( (caller(0))[0] ) unless defined $options;
 
-    my %p;
-    if ( defined $p->[0] && UNIVERSAL::isa( $p->[0], 'HASH' ) )
+    unless ( NO_VALIDATE )
     {
-	# Make a copy so we don't alter the hash reference for the
-	# caller.
-	%p = %{ $p->[0] };
+        if ( ref $p eq 'ARRAY' )
+        {
+            # we were called as validate( @_, ... ) where @_ has a
+            # single element, a hash reference
+            if ( ref $p->[0] )
+            {
+                $p = $p->[0];
+            }
+            elsif ( @$p % 2 )
+            {
+                my $called = _get_called();
+
+                $options->{on_fail}->
+                    ( "Odd number of parameters in call to $called " .
+                      "when named parameters were expected\n" );
+            }
+            else
+            {
+                $p = {@$p};
+            }
+        }
     }
-    else
-    {
-	$options->{on_fail}->
-            ( "Odd number of parameters in call to $called " .
-              "when named parameters were expected\n" )
-                if @$p % 2;
-
-	# This is to hashify the list.  Also has the side effect of
-	# copying the values so we can play with it however we want
-	# without actually changing @_.
-	%p = @$p;
-    }
-
-    if ( $ENV{PERL_NO_VALIDATION} )
-    {
-	while ( my ($key, $spec) = each %specs )
-	{
-	    if ( ! exists $p{$key} && ref $spec && exists $spec->{default} )
-	    {
-		$p{$key} = $spec->{default};
-	    }
-	}
-
-	return %p;
-    }
-
-    local $called = (caller( $options->{stack_skip} ))[3] unless defined $called;
 
     if ( $options->{ignore_case} || $options->{strip_leading} )
     {
-	%specs = _normalize_named(%specs);
-	%p = _normalize_named(%p);
+	$specs = _normalize_named($specs);
+	$p = _normalize_named($p);
+    }
+
+    if ( NO_VALIDATE )
+    {
+        return
+            ( wantarray ?
+              (
+               # this is a has containing just the defaults
+               ( map { $_ => $specs->{$_}->{default} }
+                 grep { ref $specs->{$_} && exists $specs->{$_}->{default} }
+                 keys %$specs
+               ),
+               # this recapitulates the login seen above in order to
+               # derefence our parameters properly
+               ( ref $p eq 'ARRAY' ?
+                 ( ref $p->[0] ?
+                   %{ $p->[0] } :
+                   @$p ) :
+                 %$p
+               )
+              ) :
+              do
+              {
+                  my $ref =
+                      ( ref $p eq 'ARRAY' ?
+                        ( ref $p->[0] ?
+                          $p->[0] :
+                          {@$p} ) :
+                        $p
+                      );
+
+                  foreach ( grep { ref $specs->{$_} && exists $specs->{$_}->{default} }
+                            keys %$specs )
+                  {
+                      $ref->{$_} = $specs->{$_}->{default}
+                          unless exists $ref->{$_};
+                  }
+
+                  $ref;
+              }
+            );
     }
 
     unless ( $options->{allow_extra} )
     {
-	if ( my @unmentioned = grep { ! exists $specs{$_} } keys %p )
+        my $called = _get_called();
+
+	if ( my @unmentioned = grep { ! exists $specs->{$_} } keys %$p )
 	{
 	    $options->{on_fail}->
                 ( "The following parameter" . (@unmentioned > 1 ? 's were' : ' was') .
@@ -177,32 +252,48 @@ sub validate (\@$)
     }
 
     my @missing;
-    while ( my ( $key, $spec ) = each %specs )
+ OUTER:
+    while ( my ($key, $spec) = each %$specs )
     {
-	unless ( exists $p{$key} )
-	{
-	    unless ( _is_optional($spec) )
-	    {
-		push @missing, $key;
-		next;
-	    }
+	if ( ! exists $p->{$key} &&
+             ( ref $spec ?
+               ! (
+                  do
+                  {
+                      # we want to short circuit the loop here if we
+                      # can assign a default, because there's no need
+                      # check anything else at all.
+                      if ( exists $spec->{default} )
+                      {
+                          $p->{$key} = $spec->{default};
+                          next OUTER;
+                      }
+                  }
+                  ||
+                  do
+                  {
+                      # Similarly, an optional parameter that is
+                      # missing needs no additional processing.
+                      $spec->{optional} && next OUTER
+                  }
+                 ) :
+               $spec )
+           )
+        {
+            push @missing, $key;
 	}
-
-	# Can't validate a non hashref spec beyond presence/absence of the parameter.
-	next unless ref $spec;
-
-	if ( exists $p{$key} )
-	{
-	    _validate_one_param( $p{$key}, $spec, "The '$key' parameter" );
-	}
-	else
-	{
-	    $p{$key} = $spec->{default} if exists $spec->{default};
+        # Can't validate a non hashref spec beyond the presence or
+        # absence of the parameter.
+        elsif (ref $spec)
+        {
+	    _validate_one_param( $p->{$key}, $spec, "The '$key' parameter" );
 	}
     }
 
     if (@missing)
     {
+        my $called = _get_called();
+
 	my $missing = join ', ', map {"'$_'"} @missing;
 	$options->{on_fail}->
             ( "Mandatory parameter" .
@@ -210,64 +301,43 @@ sub validate (\@$)
               " $missing missing in call to $called\n" );
     }
 
-    return %p if wantarray;
+    return wantarray ? %$p : $p;
 }
 
 sub validate_with
 {
-    my %p = validate( @_,
-                      { params => { type => ARRAYREF | HASHREF },
-                        spec   => { type => ARRAYREF | HASHREF },
-                        called =>
-                        { type => SCALAR, default => undef },
-                        ignore_case =>
-                        { type => BOOLEAN, optional => 1 },
-                        strip_leading =>
-                        { type => BOOLEAN, optional => 1 },
-                        allow_extra =>
-                        { type => BOOLEAN, optional => 1 },
-                        on_fail =>
-                        { type => CODEREF, optional => 1 },
-                        stack_skip =>
-                        { type => SCALAR, optional => 1 },
-                      }
-                    );
+    return if NO_VALIDATE && ! defined wantarray;
 
-    local $options = _get_options( (caller(0))[0] );
-    foreach my $k ( keys %$options )
+    my %p = @_;
+
+    local $options = _get_options( (caller(0))[0], %p );
+
+    unless ( NO_VALIDATE )
     {
-        $options->{$k} = $p{$k} if exists $p{$k};
-    }
+        unless ( exists $options->{called} )
+        {
+            $options->{called} = (caller( $options->{stack_skip} ))[3];
+        }
 
-    local $called = $p{called} if defined $p{called};
+    }
 
     if ( UNIVERSAL::isa( $p{spec}, 'ARRAY' ) )
     {
-	return validate_pos @{ $p{params} }, @{ $p{spec} };
+	return validate_pos( @{ $p{params} }, @{ $p{spec} } );
     }
     else
     {
-	my $params =
-            UNIVERSAL::isa( $p{params}, 'ARRAY' ) ? $p{params} : [ %{ $p{params} } ];
-
-	return validate @$params, $p{spec};
+        # intentionally ignore the prototype because this contains
+        # either an array or hash reference, and validate() will
+        # handle either one properly
+	return &validate( $p{params}, $p{spec} );
     }
 }
 
-sub _is_optional
-{
-    my $spec = shift;
-
-    # foo => 1  used to mark mandatory argument with no other validation
-    return ! $spec unless ref $spec;
-
-    # If it has a default it has to be optional
-    return exists $spec->{optional} ? $spec->{optional} : exists $spec->{default};
-}
-
 sub _normalize_named
 {
-    my %h = @_;
+    # intentional copy so we don't destroy original
+    my %h = %{ $_[0] };
 
     if ( $options->{ignore_case} )
     {
@@ -287,33 +357,37 @@ sub _normalize_named
 	}
     }
 
-    return %h;
+    return \%h;
 }
 
 sub _validate_one_param
 {
-    my $value = shift;
-    my %spec = %{ shift() };
-    my $id = shift;
+    my ($value, $spec, $id) = @_;
 
-    if ( exists $spec{type} )
+    if ( exists $spec->{type} )
     {
-	my $type = _get_type($value);
-	unless ( $type & $spec{type} )
+	unless ( _get_type($value) & $spec->{type} )
 	{
+            my $type = _get_type($value);
+
 	    my @is = _typemask_to_strings($type);
-	    my @allowed = _typemask_to_strings($spec{type});
+	    my @allowed = _typemask_to_strings($spec->{type});
 	    my $article = $is[0] =~ /^[aeiou]/i ? 'an' : 'a';
 
+            my $called = _get_called(1);
+
 	    $options->{on_fail}->
                 ( "$id to $called was $article '@is', which " .
                   "is not one of the allowed types: @allowed\n" );
 	}
     }
 
-    if ( exists $spec{isa} )
+    # short-circuit for common case
+    return unless $spec->{isa} || $spec->{can} || $spec->{callbacks};
+
+    if ( exists $spec->{isa} )
     {
-	foreach ( ref $spec{isa} ? @{ $spec{isa} } : $spec{isa} )
+	foreach ( ref $spec->{isa} ? @{ $spec->{isa} } : $spec->{isa} )
 	{
 	    unless ( UNIVERSAL::isa( $value, $_ ) )
 	    {
@@ -321,6 +395,8 @@ sub _validate_one_param
 		my $article1 = $_ =~ /^[aeiou]/i ? 'an' : 'a';
 		my $article2 = $is =~ /^[aeiou]/i ? 'an' : 'a';
 
+                my $called = _get_called(1);
+
 		$options->{on_fail}->
                     ( "$id to $called was not $article1 '$_' " .
                       "(it is $article2 $is)\n" );
@@ -328,28 +404,45 @@ sub _validate_one_param
 	}
     }
 
-    if ( exists $spec{can} )
+    if ( exists $spec->{can} )
     {
-	foreach ( ref $spec{can} ? @{ $spec{can} } : $spec{can} )
+	foreach ( ref $spec->{can} ? @{ $spec->{can} } : $spec->{can} )
 	{
-	    $options->{on_fail}->( "$id to $called does not have the method: '$_'\n" )
-		unless UNIVERSAL::can( $value, $_ );
+            unless ( UNIVERSAL::can( $value, $_ ) )
+            {
+                my $called = _get_called(1);
+
+                $options->{on_fail}->( "$id to $called does not have the method: '$_'\n" );
+            }
 	}
     }
 
-    if ( $spec{callbacks} )
+    if ( $spec->{callbacks} )
     {
-	$options->{on_fail}->
-            ( "'callbacks' validation parameter for $called must be a hash reference\n" )
-                unless UNIVERSAL::isa( $spec{callbacks}, 'HASH' );
+        unless ( UNIVERSAL::isa( $spec->{callbacks}, 'HASH' ) )
+        {
+            my $called = _get_called(1);
 
-	foreach ( keys %{ $spec{callbacks} } )
+            $options->{on_fail}->
+                ( "'callbacks' validation parameter for $called must be a hash reference\n" );
+        }
+
+
+	foreach ( keys %{ $spec->{callbacks} } )
 	{
-	    $options->{on_fail}->( "callback '$_' for $called is not a subroutine reference\n" )
-		unless UNIVERSAL::isa( $spec{callbacks}{$_}, 'CODE' );
+            unless ( UNIVERSAL::isa( $spec->{callbacks}{$_}, 'CODE' ) )
+            {
+                my $called = _get_called(1);
 
-	    $options->{on_fail}->( "$id to $called did not pass the '$_' callback\n" )
-		unless $spec{callbacks}{$_}->($value);
+                $options->{on_fail}->( "callback '$_' for $called is not a subroutine reference\n" );
+            }
+
+            unless ( $spec->{callbacks}{$_}->($value) )
+            {
+                my $called = _get_called(1);
+
+                $options->{on_fail}->( "$id to $called did not pass the '$_' callback\n" );
+            }
 	}
     }
 }
@@ -367,15 +460,13 @@ sub _validate_one_param
 
     sub _get_type
     {
-	my $value = shift;
+	return UNDEF unless defined $_[0];
 
-	return UNDEF unless defined $value;
-
-	my $ref = ref $value;
+	my $ref = ref $_[0];
 	unless ($ref)
 	{
 	    # catches things like:  my $fh = do { local *FH; };
-	    return GLOB if UNIVERSAL::isa( \$value, 'GLOB' );
+	    return GLOB if UNIVERSAL::isa( \$_[0], 'GLOB' );
 	    return SCALAR;
 	}
 
@@ -383,7 +474,7 @@ sub _validate_one_param
 
 	foreach ( keys %isas )
 	{
-	    return $isas{$_} | OBJECT if UNIVERSAL::isa( $value, $_ );
+	    return $isas{$_} | OBJECT if UNIVERSAL::isa( $_[0], $_ );
 	}
 
 	# I really hope this never happens.
@@ -444,11 +535,40 @@ sub _validate_one_param
 
     sub _get_options
     {
-	my $caller = shift;
-	return exists $OPTIONS{$caller} ? $OPTIONS{$caller} : \%defaults;
+	my ( $caller, %override ) = @_;
+
+        if ( %override )
+        {
+            return
+                ( $OPTIONS{$caller} ?
+                  { %{ $OPTIONS{$caller} },
+                    %override } :
+                  { %defaults, %override }
+                );
+        }
+        else
+        {
+            return
+                ( exists $OPTIONS{$caller} ?
+                  $OPTIONS{$caller} :
+                  \%defaults );
+        }
     }
 }
 
+sub _get_called
+{
+    my $extra_skip = $_[0] || 0;
+
+    # always add one more for this sub
+    $extra_skip++;
+
+    return ( exists $options->{called} ?
+             $options->{called} :
+             ( caller( $options->{stack_skip} + $extra_skip ) )[3]
+           );
+}
+
 1;
 
 __END__
@@ -576,7 +696,7 @@ subroutine like this:
                  ...
                } );
 
-Subroutines expected positional parameters should call the
+Subroutines expecting positional parameters should call the
 C<validate_pos> subroutine like this:
 
  validate_pos( @_, { validation spec }, { validation spec } );
@@ -796,6 +916,9 @@ Simple examples of defaults would be:
 
  my @p = validate( @_, 1, { default => 99 } );
 
+In scalar context, a hash reference or array reference will be
+returned, as appropriate.
+
 =head1 USAGE NOTES
 
 =head2 Validation failure
@@ -862,6 +985,9 @@ This is only relevant when dealing with named parameters.  If it is
 true, then the validation code will ignore the case of parameter
 names.  Defaults to false.
 
+When this is turned on, we have to copy more data around internally,
+leading to a potential speed impact.
+
 =item * strip_leading => $characters
 
 This too is only relevant when dealing with named parameters.  If this
@@ -870,6 +996,9 @@ considered equivalent to parameters without them entirely.  For
 example, if this is specified as '-', then C<-foo> and C<foo> would be
 considered identical.
 
+When this is turned on, we have to copy more data around internally,
+leading to a potential speed impact.
+
 =item * allow_extra => $boolean
 
 If true, then the validation routine will allow extra parameters not
@@ -950,6 +1079,9 @@ development but don't want the speed hit during production.
 The only error that will be caught will be when an odd number of
 parameters are passed into a function/method that expects a hash.
 
+This environment value is checked B<only> when the module is first
+loaded.  You cannot change it after the module has loaded.
+
 =head1 LIMITATIONS
 
 Right now there is no way (short of a callback) to specify that
